#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gmp.h>
#include <math.h>
#include <time.h>
#ifndef _MSC_VER
#include <pthread.h>
#include <unistd.h>
#else
#include <windows.h>
#endif


#ifdef _WIN32
#pragma warning( disable : 4127 ) /* disable conditional is constant warning */
#endif

//#define DEBUG

#define TRUE	1  // true
#define FALSE	0  //false

#define MAXPRIMES	15000000		// from 1.300.000 to 15.000.000 depending on exponent and factoring bit depth
#define MAXK		272272	 		//250000

#define NUM_FACTORING_THREADS 16

#ifndef _WIN64
#define version "5.01 - December 27th, 2007 - AMD64/Win32 compile"
#else
#define version "5.01 - December 27th, 2007 - AMD64/Win64 compile"
#endif

int	iter_last_flag = FALSE;
int sieve_folder [MAXPRIMES];		// Needed for Eratosthenes' sieve
int start_gap[17][MAXPRIMES/10];	// Fields 1..15 hold the gaps from the index of the 
									//    first field of the big matrix mod 60 for each prime,
									//    field 16 is the relative prime siever.
									// The "gap" field is updated after each sieving of the big matrix.
int big_matrix[60][MAXK];			// Contains the k values to be sieved and ordered mod60.
int mod60_classes[NUM_FACTORING_THREADS];	// Contains the allowed classes vectors) mod60 of the big matrix.
int k_mod_prime[MAXPRIMES/10];		// Contains the first k that satisfies 2kP+1 = 0

int num_primes = 0;
int num_steps  = 0;

char *results_filename = "results.txt";
char *progress_filename= "status.txt";

mpz_t one, two, powmod_ops;
mpz_t k, twop, twopplus1;
mpz_t k_val[NUM_FACTORING_THREADS],  powmod_operations[NUM_FACTORING_THREADS];

mpz_t exponent, factor, kl, k_tmp, k_max, k_act, k_act1, steps;
int start_bit, stop_bit;
int factors_found;
int factor_bits;
int factor_limbs;
int max_primes;
int limit;

int MAX_NUM_THREADS;

// A mutex is needed to synchronize the update on the results file
#ifndef _MSC_VER
pthread_mutex_t results_file_mutex = PTHREAD_MUTEX_INITIALIZER;
int rc;	//Mutex handling return code
#else
HANDLE results_file_mutex;
DWORD rc;	//Mutex handling return code
#endif

// Function prototyping
int get_k_mod_prime( int );
int lincon( int, int, int );
//float format_time( time_t, time_t );

void usage(char *program_name) {
	fprintf(stderr, "\nUsage: %s exponent bit_start bit_stop num_threads\n", program_name);
	fprintf(stderr, "\n%s --resume\n", program_name);
	fprintf(stderr, "%s --version\n", program_name);
	fprintf(stderr, "\nnum_threads must be 1, 2, 4, 8, 16\n");
	exit(EXIT_FAILURE);
}


int read_resume_file(char *filename)
{
	/* returns zero for error */
	FILE *fd;
	int i;

	fd = fopen(filename, "r");
	if (fd) { 
		gmp_fscanf(fd, "%Zu", exponent          );
		gmp_fscanf(fd, "%u", &start_bit         );
		gmp_fscanf(fd, "%u", &stop_bit          );
		gmp_fscanf(fd, "%Zu", k                 );
		gmp_fscanf(fd, "%Zu", k_tmp             );
		gmp_fscanf(fd, "%Zu", k_max             );
		gmp_fscanf(fd, "%u", &num_steps         );
		gmp_fscanf(fd, "%u", &factors_found     );
		gmp_fscanf(fd, "%u", &MAX_NUM_THREADS   );	
		gmp_fscanf(fd, "%Zu", powmod_ops );

		fclose(fd);

		for ( i = num_steps*MAX_NUM_THREADS; i < num_steps*MAX_NUM_THREADS + MAX_NUM_THREADS; i++ )
		{
			mpz_set( k_val[i], k_tmp );
		}
		return 1;
	}	
	return 0;
}


int write_resume_file( char *filename, 
	                   mpz_t exponent, 
					   int start_bit, 
					   int stop_bit, 
					   mpz_t k, 
					   mpz_t k_tmp,
					   mpz_t k_max, 
					   int num_steps,
					   int factors_found, 
					   int MAX_NUM_THREADS,
					   mpz_t powmod_ops
					 ) 
{
	/* returns zero for error */
	FILE *fd;
	
	fd = fopen(filename, "w");
	if (fd) { 
		gmp_fprintf(fd, "%Zu\n", exponent        );
		gmp_fprintf(fd, "%u\n",  start_bit       );
		gmp_fprintf(fd, "%u\n",  stop_bit        );
		gmp_fprintf(fd, "%Zu\n", k               );
		gmp_fprintf(fd, "%Zu\n", k_tmp           );
		gmp_fprintf(fd, "%Zu\n", k_max           );
		gmp_fprintf(fd, "%u\n",  num_steps       );
		gmp_fprintf(fd, "%u\n",  factors_found   );
		gmp_fprintf(fd, "%u\n",  MAX_NUM_THREADS );
		gmp_fprintf(fd, "%Zu\n", powmod_ops      );

		fflush( fd );

		fclose(fd);
		
		return 1;
	}	
	return 0;
}


#ifndef _MSC_VER
void* write_status( void* data )
#else
DWORD WINAPI write_status( void* data )
#endif
{
	mpz_t k_att, powmod_temp;
	double inizio, delta, serie, delta_div_16, completed;
	int i, qp;

	mpz_init( k_att       );
	mpz_init_set_ui( powmod_temp, 0 );

	inizio        = mpz_get_d( k );
	serie         = mpz_get_d( k_max );
	delta_div_16  = ( serie - inizio );
	delta         = delta_div_16 * 16;

#ifndef _MSC_VER
		sleep( 5 );
#else
		Sleep( 5000 );  // in milliseconds
#endif

	while ( 1 )
	{
		for ( qp = 0; qp < 10; qp++ )
		{
			mpz_set( k_att, k_val[num_steps*MAX_NUM_THREADS] );

			for ( i=num_steps*MAX_NUM_THREADS+1; i<num_steps*MAX_NUM_THREADS+MAX_NUM_THREADS; i++ )
			{
				if ( mpz_cmp( k_val[i-1], k_val[i] ) > 0 )
				{
					mpz_set( k_att, k_val[i] );
				}
			}

			completed = mpz_get_d( k_att ) - inizio;
			completed = ( completed  + num_steps * delta_div_16 ) * MAX_NUM_THREADS;
			completed = ( completed / delta ) * 100;

			gmp_printf("\nFactoring M%Zu from 2^%i to 2^%i %7.3f%% completed.", exponent, start_bit, stop_bit, completed );
			fflush( stdout );

#ifndef _MSC_VER
			sleep( 30 );
#else
			Sleep( 30000 );  // in milliseconds
#endif
		}

		write_resume_file( progress_filename, exponent, start_bit, stop_bit, k, k_att, k_max, num_steps, factors_found, MAX_NUM_THREADS, powmod_temp );	
	}

	return 0;
}

void factor_found_save(mpz_t exponent, mpz_t factor ) {
	FILE *fd;

	fd = fopen(results_filename, "a");
	if (fd) {
		gmp_fprintf(fd, "M%Zu has a factor: %Zu - Program: L5.0x\n", exponent, factor);
		fclose(fd);
	}
}

int is_merspexp(mpz_t e) {
/*
return values
	0 if e is not a Mersenne prime exponent
	1..43 position in the array otherwise
*/
	int i;
	unsigned long int mexp[] = {
	0,2,3,5,7,13,17,19,31,61,89,107,127,521,607,1279,
	2203,2281,3217,4253,4423,9689,9941,11213,19937,21701,
	23209,44497,86243,110503,132049,216091,756839,859433,
	1257787,1398269,2976221,3021377,6972593,13466917,20996011,
	24036583,25964951,30402457,32582657 };

	unsigned long int n;	
	
	if (mpz_cmp_ui(e, 2140000000) >= 0) return 0;
	n = mpz_get_ui(e);
	for (i = (sizeof mexp / sizeof *mexp) - 1; n < mexp[i]; i--)
		;
	return (n == mexp[i] ? i : 0);
}


void sieve()
{
	int i = 0;
	int j;
	int exp = 0;

	// Inizialization of variables
	memset( sieve_folder, 0, sizeof( sieve_folder ) );
	exp = mpz_fits_slong_p( exponent );
	if ( exp != 0 )
	{
		exp = mpz_get_si( exponent );
	}
	else exp = 0;

	i = 2;
	while ( i < max_primes )
	{
		j = i;
		if ( sieve_folder[j] == 0 )
		{
			if ( exp == 0 )
			{			
				start_gap[16][num_primes] = i;
				if ( i > 6 )
				{
					k_mod_prime[num_primes] = get_k_mod_prime( i );
				}
				num_primes++;
			}
			else
			{		
				if ( i != exp )
				{
					start_gap[16][num_primes] = i; 
					if ( i > 6 )
					{
						k_mod_prime[num_primes] = get_k_mod_prime( i );
					}
					num_primes++;
				}
			}

			while ( j < max_primes - i )
			{
				j = j + i;
				sieve_folder[j] = 1;
			}
		}
		i++;
	}

	return;
}


int lincon( int a, int b, int c )
// Solving the linear congruence using Euclides' algorithm
{
	int sw, m, n, p0, p1, p2, q, r, u;

	sw = 1;
	m  = b;
	n  = a;

	p0 = 1;
	p1 = m / n;
	q  = p1;
label1: 
	r = m % n;
	if ( r > 0 )
	{
		sw = -sw;
		m = n;
		n = r;
		q = m / n;
		p2 = q * p1 + p0;
		p0 = p1;
		p1 = p2;
		goto label1;
	}
	else
	{
		if ( n > 1 )
		{
			b = b / n;
			u = c / n;
			c = u;
		}
	}
	return ( sw * c * p0 % b );
}


int get_k_mod_prime( int prime )
{
	int twop_mod_prime;
	int rc;

	twop_mod_prime = mpz_fdiv_ui( twop, prime );

	// Define linear congruence for k to have 2kp+1 mod prime = 0
	rc = lincon( twop_mod_prime, prime, prime - 1 );
	if ( rc < 0)
	{
		rc = rc + prime;
	}
	return rc;
}


void create_big_matrix( mpz_t k_act1 )
{
	int i,j,w;
	mpz_t temp;

	mpz_init( temp );

	// write the k_values

	for ( i = 0; i < MAXK; i++ )
	{
		for ( j = 0; j < 60; j++ )
		{
			big_matrix[j][i] = ( j ) + ( i * 60 );
		}
	}

	// convert the value of k_tmp
	// compute the first 60 possible 2kp+1 factors

	w = 0;

	for ( i = 0; i < 60; i++ )
	{
		// create factor
		mpz_mul( temp, k_act1, twop );
		mpz_add_ui( temp, temp, 1 );

		// check it mod 60
		j = mpz_fdiv_q_ui( temp, temp, 120 );

		// build the search array and check the modulo_60 rows
		if ( ( j == 1   ) || ( j == 7   ) || ( j == 17  ) || ( j == 23 ) || 
			 ( j == 31  ) || ( j == 41  ) || ( j == 47  ) || ( j == 49 ) ||
			 ( j == 119 ) || ( j == 113 ) || ( j == 103 ) || ( j == 97 ) || 
			 ( j == 89  ) || ( j == 79  ) || ( j == 73  ) || ( j == 71 )
		   )
		{
			mod60_classes[w] = i;
			w++;
		}
		mpz_add_ui( k_act1, k_act1, 1 );
	}

	return;
}


#ifndef _MSC_VER
void* factorize( void* data )
#else
DWORD WINAPI factorize( void* data )
#endif
{
	mpz_t k_temp, prodotto, t_step;
	int k_mod_120;
	int i, w, j;
	int classe = *((int*)data);

	int working_vector[MAXK];			// Copy of the class row to sieve

	mpz_set( k_val[classe], k_act ); 
	mpz_init_set_ui( t_step, 0 );
	mpz_init( k_temp   );
	mpz_init( prodotto ); 
	// time_t start,stop;

	// Look for actual k mod 120
	mpz_add_ui( k_temp, k_val[classe], mod60_classes[classe] );
	mpz_mul( k_temp, k_temp, twop );
	mpz_add_ui( k_temp, k_temp, 1 );

#ifdef DEBUG
	printf( "\n Starting thread %2i: ", classe );
	fflush( stdout );
#endif

	// lincon1 vars
	mpz_t sw, m, p0, p1, p2, q, r, u;
	mpz_t n, c1, b1;

	mpz_init( sw ); 
	mpz_init( m );
	mpz_init( b1 );
	mpz_init( c1 );
	mpz_init( n );
	mpz_init( p0 );
	mpz_init( p1 );
	mpz_init( p2 );
	mpz_init( q );
	mpz_init( r );
	mpz_init( u );
	mpz_init_set_ui( powmod_operations[classe], 0 );

	mpz_mul_ui(prodotto, exponent, 120 );
	
	for ( i = 3; i < num_primes; i++ )
	{
		k_mod_120 = mpz_fdiv_ui( k_temp, start_gap[16][i] );

		//start_gap[classe][i] = lincon1( prodotto, start_gap[16][i], ( - k_mod_120 ) );
		
		// Solving the linear congruence using Euclides' algorithm

		mpz_set_si( sw, 1); 
		mpz_set_si( m, start_gap[16][i]);
		mpz_set_si( b1, start_gap[16][i] );
		mpz_set_si( c1, ( - k_mod_120 ) );
		mpz_set( n, prodotto );

		mpz_set_si( p0, 1);
		mpz_tdiv_q( p1, m, n );
		mpz_set( q, p1 );
	label1: 
		mpz_fdiv_r( r, m, n );
		if ( mpz_cmp_si( r, 0 ) > 0 )
		{
			mpz_neg( sw, sw);
			mpz_set( m, n);
			mpz_set( n, r);
			mpz_tdiv_q( q, m, n);
			//p2 = q * p1 + p0;
			mpz_mul( p2, q, p1 );
			mpz_add( p2, p2, p0 );
			mpz_set( p0, p1);
			mpz_set( p1, p2);
			goto label1;
		}
		else
		{
			if ( mpz_cmp_ui( n, 1 ) > 0 ) // > 1 ?
			{
				mpz_tdiv_q( b1, b1, n );
				mpz_tdiv_q( u, c1, n);
				mpz_set( c1, u );
			}
		}
		mpz_mul( sw, sw, c1 );
		mpz_mul( sw, sw, p0 );
		mpz_fdiv_r( sw, sw, b1 );
		start_gap[classe][i] = mpz_get_si( sw );
		// Insert the following line to look for very small factors, < max_prime
		//if ( ( start_gap[classe][i] < 0 ) || ( mpz_cmp_ui( twopplus1, start_gap[classe][i] ) <= 0 ) ) 
		if ( ( start_gap[classe][i] < 0 ) )
		{
			start_gap[classe][i] = start_gap[classe][i] + start_gap[16][i];	
		}
	}

	// Start sieving routine

	//i = 0;
	while ( mpz_cmp( steps, t_step ) >= 0 ) 
	{	
		// Copy a row from the big_matrix to a modifiable workspace
		for ( w = 0; w < MAXK; w++ )
		{
			working_vector[w] = big_matrix[mod60_classes[classe]][w];
		}

		// Set to -1 divisors on the bitmap
		for ( w = 3; w < num_primes; w++ )
		{
			j = start_gap[classe][w];

			while ( j < MAXK )
			{
				working_vector[j] = -1;
				j = j + start_gap[16][w];
			}		
			
			// Update gap table - j > MAXK
			j = j - MAXK;					
			start_gap[classe][w] = j;
		}

		// Start powmod routine
		for ( w = 0; w < MAXK; w++ )
		{
			if ( working_vector[w] != -1)
			{
				// compute 2kp+1
				mpz_add_ui( k_temp, k_val[classe], working_vector[w] );
				mpz_mul( k_temp, k_temp, twop );
				mpz_add_ui( k_temp, k_temp, 1 );

				mpz_powm( prodotto, two, exponent, k_temp );

			   if ( ( ( mpz_cmp_ui( prodotto, 1 ) ) == 0 ) && ( mpz_probab_prime_p( k_temp, 10 ) > 0 ) )
			   { 
					// Lock the mutex to write  file
#ifndef _MSC_VER
					rc = pthread_mutex_lock( &results_file_mutex );
					if (rc) { /* an error has occurred */
						perror("pthread_mutex_lock");
						pthread_exit(NULL);
					}
#else
					rc = WaitForSingleObject( results_file_mutex, INFINITE );
					if (rc == WAIT_FAILED) { /* an error has occurred */
						printf("factorize(): WaitForSingleObject(): MUTEX Failure\n");
						return 0;
					}
#endif

					printf( "\nM" ); gmp_printf("%Zd",exponent); printf(" has a factor: "); gmp_printf( "%Zd - Program: L5.0x", k_temp );
					fflush( stdout );
					factors_found++;
					factor_found_save(exponent, k_temp );

					// Unlock the mutex to write  file
#ifndef _MSC_VER
					rc = pthread_mutex_unlock( &results_file_mutex );
					if (rc) { /* an error has occurred */
						perror("pthread_mutex_unlock");
						pthread_exit(NULL);
					}
#else
					rc = ReleaseMutex( results_file_mutex );
					if (rc == 0) { /* an error has occurred */
						printf("factorize(): ReleaseMutex: MUTEX Failure\n");
						return 0;
					}
#endif
			   } 

				mpz_add_ui( powmod_operations[classe], powmod_operations[classe], 1 );
			}
		}
		// update k
		mpz_add_ui( k_val[classe], k_val[classe], MAXK * 60 );
		mpz_add_ui( t_step, t_step, 1 );
	} 

#ifdef DEBUG
	printf( "\n Stopping thread %2i: ", classe );
	fflush( stdout );
#endif

	return 0;
}


int main ( int argc, char *argv[] ) 
{
	FILE *fd;
	mpz_t number_start, number_stop, gmp_start_bit, gmp_stop_bit, gmp_num_threads, zero;
	// int result = 0;
	int f,g;
	time_t start_date, end_date;
	void* thr_retval[17] = { 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 };
	int thr_id[NUM_FACTORING_THREADS+1];
#ifndef _MSC_VER
	pthread_t p_thread[NUM_FACTORING_THREADS+1];
#else
	HANDLE hThreads[NUM_FACTORING_THREADS+1];

	results_file_mutex = CreateMutex( NULL, FALSE, NULL );
#endif

	mpz_init( gmp_start_bit        );
	mpz_init( gmp_stop_bit         );
	mpz_init( gmp_num_threads      );
	mpz_init( number_stop          );
	mpz_init( number_start         );
	mpz_init( k                    );
	mpz_init( kl                   );
	mpz_init( k_tmp                );
	mpz_init( k_max                );
	mpz_init( k_act                );
	mpz_init( k_act1               );
	mpz_init( steps                );
	mpz_init( exponent             );
	mpz_init( factor               );
	mpz_init( twop                 );
	mpz_init( twopplus1            );
	mpz_init_set_ui( powmod_ops, 0 );
	mpz_init_set_ui	( zero,      0 );
	mpz_init_set_ui( one,        1 );
	mpz_init_set_ui( two,        2 );

	for ( f = 0; f < 16; f++ )
	{
		mpz_init( k_val[f] );
	}

	// Command line parsing
	/* read command line arguments */

	if ( ( argc != 2 ) && ( argc != 5 ) )
	{
		usage(argv[0]);
		exit(EXIT_FAILURE);
	}
	if (argc == 2) {
		if (!strcmp(argv[1], "--resume")) {
			if (!read_resume_file(progress_filename ) ) {
				fprintf(stderr, "error reading %s\n", progress_filename);
				exit(EXIT_FAILURE);
			}
			gmp_printf( "\nResuming factoring of M%Zu from 2^%i to 2^%i", exponent, start_bit, stop_bit );

			mpz_mul_ui( twop, exponent, 2 );
			mpz_add_ui( twopplus1, twop, 1 );
			factor_bits  = ( int )mpz_sizeinbase( twopplus1, 2 ) + stop_bit;
			factor_limbs = factor_bits / 32;
			limit = 15000000 / ( factor_limbs * MAX_NUM_THREADS * 300);

			// k values normalization
			// kl holds the range of k needed to reach number_stop

			mpz_sub( kl, k_max, k_tmp );

			// Define number of steps 

			mpz_tdiv_q_ui( steps, kl, ( MAXK * 60 ) );
		
			if ( mpz_cmp_ui( steps, limit ) > 0 )
			{
				max_primes = MAXPRIMES;
			}
			else
			{
				max_primes = ( factor_limbs * MAX_NUM_THREADS * 300 * ( mpz_get_ui( steps ) + 1 ) ) + 1300000;
			}

			gmp_printf("\nFactor5 v. %s - GMP %s\n",version, gmp_version);
			fflush(stdout);
			time(&end_date);
			gmp_printf("\nCurrent date %s", ctime(&end_date));

			mpz_set( k_act, k_tmp );
			mpz_set( k_act1, k_tmp );
			create_big_matrix( k_act1 );

#ifndef _MSC_VER
			thr_id[16] = pthread_create(&p_thread[16], NULL, write_status, (void*)16);
#else
			hThreads[16] = CreateThread(NULL, 0, write_status, (void*)16, 0, NULL);
#endif

			// Sieve

			sieve();

			f = num_steps * MAX_NUM_THREADS;

			for ( g = f; g < f + MAX_NUM_THREADS; g++ )
			{
				// factorize( i, steps );
				thr_id[g] = g;
#ifndef _MSC_VER
				pthread_create( &p_thread[g], NULL, factorize, (void*)&thr_id[g] );
#else
				hThreads[g] = CreateThread(NULL, 0, factorize, (void*)&thr_id[g], 0, NULL);
#endif
			}

			for ( g = f; g < f + MAX_NUM_THREADS; g++ )
			{
				// Wait until threads stop
				// void* thr_retval;
#ifndef _MSC_VER
				pthread_join( p_thread[g], thr_retval[g] );
#else
				WaitForSingleObject(hThreads[g], INFINITE);
				CloseHandle(hThreads[g]);
#endif
			}
			num_steps++;
		} 
		else if (!strcmp(argv[1], "--version")) 
		{
			gmp_printf("\nFactor5 v. %s - GMP %s\n",version, gmp_version);
			fflush(stdout);
			exit(EXIT_SUCCESS);
		} 
		else 
		{
			usage(argv[0]);
			exit(EXIT_FAILURE);
		}
	} 
	else if (argc == 5) 
	{
		// The following 5 lines should be commented to run batch files
		// ------------------
		if (read_resume_file( progress_filename ) ) 
		{
			fprintf(stderr, "use --resume or delete %s\n", progress_filename);
			exit(EXIT_FAILURE);
		}
		// -------------------

		mpz_set_str(exponent, argv[1], 10);
		mpz_set_str(gmp_start_bit, argv[2], 10);
		mpz_set_str(gmp_stop_bit, argv[3], 10);
		mpz_set_str(gmp_num_threads, argv[4], 10);
		time(&start_date);
		factors_found = 0;

		/* check arguments */
		
		if (!mpz_probab_prime_p(exponent, 10)) {
			fprintf(stderr, "exponent not a prime number\n");
			return EXIT_FAILURE;
		}
		if (is_merspexp(exponent)) {
			fprintf(stderr, "Mersenne prime exponent\n");
			return EXIT_FAILURE;
		}

		if ( mpz_cmp( gmp_start_bit, gmp_stop_bit ) >= 0 ) {
			fprintf(stderr, "start_bit >= stop_bit\n");
			return EXIT_FAILURE;
		}

		start_bit = mpz_get_ui( gmp_start_bit );
		stop_bit = mpz_get_ui( gmp_stop_bit );
		MAX_NUM_THREADS = mpz_get_ui( gmp_num_threads );

		if ( ( MAX_NUM_THREADS != 1 ) && ( MAX_NUM_THREADS != 2 ) && ( MAX_NUM_THREADS != 4 ) && ( MAX_NUM_THREADS != 8 ) && ( MAX_NUM_THREADS != 16 ) )
		{
			usage( argv[0] );
		}

		mpz_mul_ui( twop, exponent, 2 );
		mpz_add_ui( twopplus1, twop, 1 );
		factor_bits  = ( int )mpz_sizeinbase( twopplus1, 2 ) + stop_bit;
		factor_limbs = factor_bits / 32;
		limit = 15000000 / ( factor_limbs * MAX_NUM_THREADS * 300);

		/* override number_start if resuming */
		if (mpz_cmp_ui(factor, 0) > 0) {
			mpz_set(number_start, factor);
		} else {
			/* number_start = 2^bit_start or the smallest factor of the form 2kp + 1*/
			/* 2p + 1*/
			mpz_set(number_start, one);
			mpz_mul_2exp(number_start, number_start, start_bit);
			if (mpz_cmp( number_start, twopplus1 ) < 0 )
			{
				mpz_set( number_start, twopplus1 );
				mpz_set_ui( k, 1 );
			}
			else // Let's compute the starting k 
			{
				mpz_sub_ui( kl, number_start, 1 );
				mpz_tdiv_q( k, kl, twop );
			}
		}

		/* number_stop = 2^bit_stop - 1  */
		mpz_set(number_stop, one);
		mpz_mul_2exp(number_stop, number_stop, stop_bit);
		mpz_sub(number_stop, number_stop, one);

		if (mpz_cmp( number_start, number_stop ) > 0)
		{
			fprintf(stderr, "stop_bit too low!\n");
			return EXIT_FAILURE;
		}

		// Now let's compute the maximum k from number_stop

		mpz_sub_ui( kl, number_stop, 1 );
		mpz_cdiv_q( k_max, kl, twop );

		fd = fopen(results_filename, "a");
		if (fd) {
			gmp_fprintf(fd, "Trial-factoring M%Zu in [2^%i, 2^%i-1]\n", exponent, start_bit, stop_bit );
			fclose(fd);
		}

		gmp_printf("\nFactor5 v. %s - GMP %s\n",version, gmp_version);
		fflush(stdout);
		time(&end_date);
		gmp_printf("\nCurrent date %s", ctime(&end_date));

#ifndef _MSC_VER
	    thr_id[16] = pthread_create(&p_thread[16], NULL, write_status, (void*)16);
#else
		hThreads[16] = CreateThread(NULL, 0, write_status, (void*)16, 0, NULL);
#endif
	}

	// k values normalization
	// kl holds the range of k needed to reach number_stop

	mpz_sub( kl, k_max, k );

	// Define number of steps 

	mpz_tdiv_q_ui( steps, kl, ( MAXK * 60 ) );

#ifdef DEBUG
	gmp_printf( "\nsteps = %Zu\n", steps );
#endif
	if ( mpz_cmp_ui( steps, limit ) > 0 )
	{
		max_primes = MAXPRIMES;
	}
	else
	{
		max_primes = ( factor_limbs * MAX_NUM_THREADS * 300 * ( mpz_get_ui( steps ) + 1 ) ) + 1300000;
	}

	// Create big matrix

	mpz_set( k_act, k );
	mpz_set( k_act1, k );

	create_big_matrix( k_act1 );

	// Sieve
	num_primes = 0;
	sieve();	

	for ( f = num_steps; f < NUM_FACTORING_THREADS / MAX_NUM_THREADS; f++ )
	{
		for ( g = f*MAX_NUM_THREADS; g < f*MAX_NUM_THREADS + MAX_NUM_THREADS; g++ )
		{
			// factorize( i, steps );
			thr_id[g] = g;
#ifndef _MSC_VER
			pthread_create( &p_thread[g], NULL, factorize, (void*)&thr_id[g] );
#else
			hThreads[g] = CreateThread(NULL, 0, factorize, (void*)&thr_id[g], 0, NULL);
#endif
		}
		for ( g = f*MAX_NUM_THREADS; g < f*MAX_NUM_THREADS + MAX_NUM_THREADS; g++ )
		{
			// Wait until threads stop
			// void* thr_retval[g];
#ifndef _MSC_VER
			pthread_join( p_thread[g], thr_retval[g] );
#else
			WaitForSingleObject(hThreads[g], INFINITE);
			CloseHandle(hThreads[g]);
#endif
		}
		num_steps++;
	}

#ifndef _MSC_VER
	rc = pthread_mutex_destroy( &results_file_mutex );
#else
    CloseHandle( results_file_mutex );
#endif

	
	// Statistics

	switch ( factors_found )
	{
		case 0 : printf("\nNo factor "); break;
		case 1 : printf("\n1 factor "); break;
		default: printf("\n%i factors ", factors_found ); break;
	}
	printf("found\n");

	for ( f=0; f<NUM_FACTORING_THREADS; f++ )
	{
		mpz_add( powmod_ops, powmod_ops, powmod_operations[f] );
	}

	gmp_printf( "\nPerformed %12Zd powmod operations since last restart.", powmod_ops );
	fflush( stdout );
	printf( "\nUsed %i primes, max. prime = %i", num_primes, start_gap[16][num_primes - 1] );
	fflush( stdout );
	
	fd = fopen(results_filename, "a");
	if (fd) {
		gmp_fprintf(fd, "M%Zu has %i factors in [2^%i, 2^%i-1].\n", exponent, factors_found, start_bit, stop_bit );
		fclose(fd);
	}
	
	time(&end_date);
	gmp_printf("\nCurrent date %s", ctime(&end_date));

	return 0;
}
